import euclidlib as el


def test_correlation_functions(data_path):
    xis = el.photo.correlation_functions(
        data_path / "two_point_correlation_function_Xi.fits"
    )

    expected_keys = [
        ("SHE", "SHE", 2, 2),
        ("SHE", "SHE", 2, 5),
        ("SHE", "SHE", 5, 5),
    ]

    assert list(xis.keys()) == expected_keys



from __future__ import annotations

from os import PathLike

import fitsio  # type: ignore [import-not-found]
from numpy.typing import NDArray

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import TypeAlias, Any

    _DictKey: TypeAlias = str | int | tuple["_DictKey", ...]


def _key_from_string(s: str) -> tuple[str, str, int, int] | None:
    """return key from string representation"""
    tuple_key_dict = None
    if "POSPOS" in s:
        name_split = s.split("_")
        tuple_key_dict = ("POS", "POS", int(name_split[-2]), int(name_split[-1]))
    elif "POSSHEAR" in s:
        name_split = s.split("_")
        tuple_key_dict = ("POS", "SHE", int(name_split[-2]), int(name_split[-1]))
    elif "SHEAR" in s:
        name_split = s.split("_")
        tuple_key_dict = ("SHE", "SHE", int(name_split[-2]), int(name_split[-1]))
    return tuple_key_dict


def correlation_functions(path: str | PathLike[str]) -> dict[_DictKey, NDArray[Any]]:
    """
    Reads 2D correlation functions from a Euclid data product.

    Parameters
    ----------
    path : str | PathLike[str]
        The path to the FITS file containing the Euclid data product.

    Returns
    -------
    dict[_DictKey, NDArray[Any]]
        A dictionary where each key is a tuple representing the HDU name (generated by
        `_key_from_string`), and each value is a NumPy array containing the data of
        the corresponding HDU.

    Notes
    -----
    - Only HDUs whose extension name contains '2D' are considered.
    - The `fitsio.FITS` context manager is used to read the FITS file efficiently.
    - The function depends on `_key_from_string()` to convert the HDU names into keys
      for the output dictionary.
    """

    xi: dict[_DictKey, NDArray[Any]] = {}
    with fitsio.FITS(path) as fits:
        for hdu in fits:
            if "2D" not in hdu.get_extname():
                continue
            key = _key_from_string(hdu.get_extname())
            if key is None:
                continue
            xi[key] = hdu.read()
    return xi


def bandpowers(path: str | PathLike[str]) -> dict[_DictKey, NDArray[Any]]:
    """
    Reads 2D bandpowers from a Euclid data product.

    Parameters
    ----------
    path : str | PathLike[str]
        The path to the FITS file containing the Euclid bandpower data product.

    Returns
    -------
    dict[_DictKey, NDArray[Any]]
        A dictionary where each key is a tuple representing the HDU name (generated by
        `_key_from_string`), and each value is a NumPy array containing the data of
        the corresponding HDU.

    Notes
    -----
    - Only HDUs whose extension name contains '2D' are considered.
    - The `fitsio.FITS` context manager is used to read the FITS file efficiently.
    - The function depends on `_key_from_string()` to convert the HDU names into keys
      for the output dictionary.
    """
    bandp: dict[_DictKey, NDArray[Any]] = {}
    with fitsio.FITS(path) as fits:
        for hdu in fits:
            if "2D" not in hdu.get_extname():
                continue
            key = _key_from_string(hdu.get_extname())
            if key is None:
                continue
            data = hdu.read()
            if key[:2] == ("POS", "POS"):
                data.dtype.names = ["L", "CL", "LMIN", "LMAX"]
            elif key[:2] == ("POS", "SHE"):
                data.dtype.names = ["L", "CL_E", "CL_B", "LMIN", "LMAX"]
            elif key[:2] == ("SHE", "SHE"):
                data.dtype.names = ["L", "CL_E", "CL_B", "LMIN", "LMAX"]
            bandp[key] = data
    return bandp


def cosebis(path: str | PathLike[str]) -> dict[_DictKey, NDArray[Any]]:
    """
    Reads 2D bandpowers from a Euclid data product.

    Parameters
    ----------
    path : str | PathLike[str]
        The path to the FITS file containing the Euclid bandpower data product.

    Returns
    -------
    dict[_DictKey, NDArray[Any]]
        A dictionary where each key is a tuple representing the HDU name (generated by
        `_key_from_string`), and each value is a NumPy array containing the data of
        the corresponding HDU.

    Notes
    -----
    - Only HDUs whose extension name contains '2D' are considered.
    - The `fitsio.FITS` context manager is used to read the FITS file efficiently.
    - The function depends on `_key_from_string()` to convert the HDU names into keys
      for the output dictionary.
    """

    cb: dict[_DictKey, NDArray[Any]] = {}
    with fitsio.FITS(path) as fits:
        for hdu in fits:
            if "2D" not in hdu.get_extname():
                continue
            key = _key_from_string(hdu.get_extname())
            if key is None:
                continue
            cb[key] = hdu.read()
    return cb



__all__ = [
    "__version__",
    "__version_tuple__",
    "photo",
    "spectro",
]

# generated version information
try:
    from ._version import __version__, __version_tuple__  # type: ignore [import-not-found, unused-ignore]
except ModuleNotFoundError:
    pass

# import submodules here so everything is available with package import
from . import photo, spectro



from __future__ import annotations

import re

import fitsio  # type: ignore [import-not-found]
import numpy as np

from dataclasses import dataclass

TYPE_CHECKING = False
if TYPE_CHECKING:
    from os import PathLike
    from typing import Any, TypeAlias
    from numpy.typing import NDArray

    _DictKey: TypeAlias = str | int | tuple["_DictKey", ...]


def normalize_result_axis(
    axis: tuple[int, ...] | int | None,
    result: NDArray[Any],
    ell: tuple[NDArray[Any], ...] | NDArray[Any] | None,
) -> tuple[int, ...]:
    """Return an axis tuple for a result."""
    try:
        from numpy.lib.array_utils import normalize_axis_tuple
    except ModuleNotFoundError:
        from numpy.lib.stride_tricks import normalize_axis_tuple  # type: ignore

    if axis is None:
        if result.ndim == 0:
            axis = ()
        elif isinstance(ell, tuple):
            axis = tuple(range(-len(ell), 0))
        else:
            axis = -1
    return normalize_axis_tuple(axis, result.ndim, "axis")


@dataclass(frozen=True, repr=False)
class Result:
    """
    Container for results.
    """

    array: NDArray[Any]
    ell: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    axis: int | tuple[int, ...] | None = None
    lower: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    upper: NDArray[Any] | tuple[NDArray[Any], ...] | None = None
    weight: NDArray[Any] | tuple[NDArray[Any], ...] | None = None

    def __post_init__(self) -> None:
        # Ensure array is of float dtype
        float_array = np.asarray(self.array, dtype=float)
        object.__setattr__(self, "array", float_array)

        # Normalize the axis after setting the array
        axis = normalize_result_axis(self.axis, self.array, self.ell)
        object.__setattr__(self, "axis", axis)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(axis={self.axis!r})"

    def __array__(
        self,
        dtype: np.dtype[Any] | None = None,
        *,
        copy: np.bool[bool] | None = None,
    ) -> NDArray[Any]:
        if copy is not None:
            return self.array.__array__(dtype, copy=copy)
        return self.array.__array__(dtype)

    def __getitem__(self, key: Any) -> Any:
        return self.array[key]

    @property
    def ndim(self) -> int:
        return self.array.ndim

    @property
    def shape(self) -> tuple[int, ...]:
        return self.array.shape

    @property
    def dtype(self) -> np.dtype[Any]:
        return self.array.dtype


def _key_from_string(s: str) -> _DictKey:
    """
    Return key for a given string representation.
    """
    parts = re.split(r"(?<!\\)-", s.replace("\\\\", "\0"))
    if len(parts) > 1:
        return tuple(map(_key_from_string, parts))
    key = parts[0]
    key = key.replace("\\-", "-")
    key = key.replace("\0", "\\")
    return int(key) if key.removeprefix("-").isdigit() else key


def _read_metadata(hdu: Any) -> dict[str, Any]:
    """read array metadata from FITS HDU"""
    h = hdu.read_header()
    md = {}
    for key in h:
        if key.startswith("META "):
            md[key[5:].lower()] = h[key]
    return md


def _read_result(hdu: Any) -> Result:
    """
    Read a result array from FITS.
    """

    from ast import literal_eval

    # read columnar data from extension
    data = hdu.read()
    h = hdu.read_header()

    # the angular axis
    axis = literal_eval(h["ELLAXIS"])

    # get data array and move axis back to right position
    arr = np.moveaxis(data["ARRAY"], tuple(range(len(axis))), axis)

    # sort ell axes into natural order
    order = np.argsort(axis)

    # get ells
    _ell = data["ELL"]
    if _ell.ndim == 1:
        ell = _ell
    else:
        ell = tuple(_ell[: arr.shape[axis[i]], i] for i in order)

    # get lower bounds
    _lower = data["LOWER"]
    if _lower.ndim == 1:
        lower = _lower
    else:
        lower = tuple(_lower[: arr.shape[axis[i]], i] for i in order)

    # get upper bounds
    _upper = data["UPPER"]
    if _upper.ndim == 1:
        upper = _upper
    else:
        upper = tuple(_upper[: arr.shape[axis[i]], i] for i in order)

    # get weights
    _weight = data["WEIGHT"]
    if _weight.ndim == 1:
        weight = _weight
    else:
        weight = tuple(_weight[: arr.shape[axis[i]], i] for i in order)

    # construct result array with ancillary arrays and metadata
    return Result(
        arr.view(np.dtype(arr.dtype, metadata=_read_metadata(hdu))),
        axis=tuple(axis[i] for i in order),
        ell=ell,
        lower=lower,
        upper=upper,
        weight=weight,
    )


def read(path: str | PathLike[str]) -> dict[_DictKey, Result]:
    """
    Read a set of results from a FITS file.
    """

    # the returned set of cls
    results = {}

    # read all HDUs in file into dict keys
    with fitsio.FITS(path) as fits:
        for hdu in fits:
            if not hdu.has_data():
                continue
            ext = hdu.get_extname()
            if not ext:
                continue
            key = _key_from_string(ext)
            if not key:
                continue
            results[key] = _read_result(hdu)

    return results


def angular_power_spectra(path: str | PathLike[str]) -> dict[_DictKey, Result]:
    """
    Read a set of results from a FITS file.
    """

    # the returned set of cls
    results = read(path)
    return results


def mixing_matrices(path: str | PathLike[str]) -> dict[_DictKey, Result]:
    """
    Read a set of results from a FITS file.
    """

    # the returned set of mixing matrices
    results = read(path)
    return results



"""
Module for internal utility functions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, TypeVar

    AnyT = TypeVar("AnyT", bound=Any)


def writer(func: Any) -> Callable[[AnyT], AnyT]:
    """
    Decorator for writer functions.
    """

    def decorator(writer: AnyT) -> AnyT:
        func.write = writer
        writer.__module__ = func.__module__
        writer.__name__ = "write"
        writer.__qualname__ = f"{func.__qualname__}.write"
        return writer

    return decorator



import numpy as np  # type: ignore
import pytest  # type: ignore

import euclidlib as el


@pytest.mark.parametrize("read_hist", [True, False])
@pytest.mark.parametrize("write_hist", [True, False])
def test_redshift_distributions(tmp_path, write_hist, read_hist):
    def fn(z, hist):
        """produce a mock n(z), optionally binned into a histogram"""
        nz = np.exp(-((z - 1.0) ** 2) / (0.5) ** 2 / 2)
        if hist:
            nz = (nz[:-1] + nz[1:]) / 2 * np.diff(z)
        return nz

    # write test data from higher redshift resolution
    z = np.linspace(0.0, 6.0, 10001)
    nz = fn(z, write_hist)

    path = tmp_path / "nz.fits"

    el.photo.redshift_distributions.write(path, z, nz, hist=write_hist)

    # read test data

    z, nz = el.photo.redshift_distributions(path, hist=read_hist)

    np.testing.assert_array_equal(z, np.linspace(0.0, 6.0, 3001))

    assert len(nz) == 1
    assert list(nz.keys()) == [1]

    nz_ = fn(z, read_hist)
    np.testing.assert_allclose(nz[1], nz_, rtol=1e-2, atol=1e-4)


def test_redshift_distributions_ident(tmp_path):
    """
    Test that writing a histogram in the correct format returns the data
    unchanged.
    """
    z_ = np.linspace(0.0, 6.0, 3001)
    zmid = (z_[:-1] + z_[1:]) / 2
    nz_ = np.exp(-((zmid - 1.0) ** 2) / (0.5) ** 2 / 2)

    path = tmp_path / "nz.fits"

    el.photo.redshift_distributions.write(path, z_, nz_, hist=True)
    z, nz = el.photo.redshift_distributions(path, hist=True)

    np.testing.assert_array_equal(z, z_)
    np.testing.assert_array_equal(nz[1], nz_.astype(nz[1].dtype))



from __future__ import annotations

from collections.abc import Mapping
from os import PathLike
from typing import TYPE_CHECKING

import fitsio  # type: ignore [import-not-found]
import numpy as np

from .._util import writer

if TYPE_CHECKING:
    from typing import Any
    from numpy.typing import ArrayLike, NDArray

if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
    trapezoid = np.trapezoid
else:
    trapezoid = np.trapz  # type: ignore


def _hist2dist(x: NDArray[Any], y: NDArray[Any]) -> NDArray[Any]:
    """
    Convert histogram to distribution.
    """
    *dims, n = y.shape
    z = np.zeros((*dims, n + 1), dtype=y.dtype.base)
    np.cumsum(y, axis=-1, out=z[..., 1:])
    result: NDArray[Any] = np.gradient(z, x, axis=-1, edge_order=1)
    return result


def redshift_distributions(
    path: str | PathLike[str],
    *,
    ext: str | int | None = None,
    hist: bool = False,
) -> tuple[NDArray[Any], Mapping[int, NDArray[Any]]]:
    """Read redshift distributions in Euclid format.

    Parameters
    ----------
    path : str
        Path to a FITS file in Euclid format.
    ext : str or int or None, optional
        The FITS extension to read.  If ``None``, the first extension
        with data is used.
    hist : bool, optional
        By default, the histograms are converted to distributions.  If
        true, return the redshift histograms unmodified.

    Returns
    -------
    z : ndarray
        Redshift values.
    nz: dict of int and ndarray
        Dictionary where keys are tomographic bin IDs and values are the
        redshift distributions.

    """

    # data and header from file
    data = fitsio.read(path, ext=ext, lower=True)

    # this is the fixed binning scheme used by PHZ
    z = np.linspace(0.0, 6.0, 3001)

    # check format
    shape = (z.size - 1,)
    if "n_z" not in data.dtype.fields or data.dtype.fields["n_z"][0].shape != shape:
        msg = f"{path}: requires column N_Z of shape {shape}"
        raise ValueError(msg)

    # read each n(z) histogram into a dict
    out = {}
    for row in data:
        bin_id, n_z = row["bin_id"], row["n_z"]
        if not hist:
            n_z = _hist2dist(z, n_z)
        out[bin_id] = n_z

    return z, out


@writer(redshift_distributions)
def _(
    path: str | PathLike[str],
    z: ArrayLike,
    nz: ArrayLike,
    *,
    weight_method: str = "NO_WEIGHT",
    bin_type: str = "TOM_BIN",
    hist: bool = False,
) -> None:
    """
    Write n(z) data in Euclid SGS format.  Supports both distributions
    (when *hist* is false, the default) and histograms (when *hist* is
    true).
    """

    z = np.asanyarray(z)
    nz = np.asanyarray(nz)

    if z.ndim != 1:
        raise ValueError("z array must be 1D")
    if nz.ndim == 0:
        raise ValueError("nz array must be at least 1D")
    if not hist and z.shape[-1] == nz.shape[-1]:
        pass
    elif hist and z.shape[-1] == nz.shape[-1] + 1:
        pass
    else:
        raise ValueError("shape mismatch between redshifts and values")

    # PHZ uses a fixed binning scheme with z bins in [0, 6] and dz=0.002
    zbinedges = np.linspace(0.0, 6.0, 3001)

    # turn nz into a 2D array with NBIN rows
    nz = nz.reshape(-1, nz.shape[-1])
    nbin = nz.shape[0]

    # create the output data in the correct format
    out = np.empty(
        nbin,
        dtype=[
            ("BIN_ID", ">i4"),
            ("MEAN_REDSHIFT", ">f4"),
            ("N_Z", ">f4", (3000,)),
        ],
    )

    # set increasing bin IDs
    out["BIN_ID"] = np.arange(1, nbin + 1)

    # convert every nz into the PHZ format
    if hist:
        # rebin the histogram as necessary

        # shorthand for the left and right z boundaries, respectively
        zl, zr = z[:-1], z[1:]

        # compute the mean redshifts
        out["MEAN_REDSHIFT"] = np.sum((zl + zr) / 2 * nz, axis=-1) / np.sum(nz, axis=-1)

        # compute resummed bin counts
        for j, (z1, z2) in enumerate(zip(zbinedges, zbinedges[1:])):
            frac = (np.clip(z2, zl, zr) - np.clip(z1, zl, zr)) / (zr - zl)
            out["N_Z"][:, j] = np.dot(nz, frac)
    else:
        # integrate the n(z) over each histogram bin

        # compute mean redshifts
        out["MEAN_REDSHIFT"] = trapezoid(z * nz, z, axis=-1) / trapezoid(nz, z, axis=-1)

        # compute the combined set of z grid points from data and binning
        zp = np.union1d(z, zbinedges)

        # integrate over each bin
        for i in range(nbin):
            # interpolate dndz onto the unified grid
            nzp = np.interp(zp, z, nz[i], left=0.0, right=0.0)

            # integrate the distribution over each bin
            for j, (z1, z2) in enumerate(zip(zbinedges, zbinedges[1:])):
                sel = (z1 <= zp) & (zp <= z2)
                out["N_Z"][i, j] = trapezoid(nzp[sel], zp[sel])

    # metadata
    header = {
        "WEIGHT_METHOD": weight_method,
        "BIN_TYPE": bin_type,
        "NBIN": nbin,
    }

    # write output data to FITS
    with fitsio.FITS(path, "rw", clobber=True) as fits:
        fits.write(out, extname="BIN_INFO", header=header)



import euclidlib as el


def test_writable():
    def func():
        pass

    func.__module__ = "mymodule"

    @el._util.writer(func)
    def writer():
        pass

    assert func.write is writer
    assert func.write.__module__ == "mymodule"
    assert func.write.__name__ == "write"
    assert func.write.__qualname__ == f"{func.__qualname__}.write"



import euclidlib as el


def test_angular_power_spectra(data_path):
    cls = el.photo.angular_power_spectra(data_path / "angular_power_spectra.fits")

    expected_keys = {
        ("POS", "POS", 2, 2),
        ("POS", "POS", 2, 5),
        ("POS", "POS", 5, 5),
        ("SHE", "SHE", 2, 2),
        ("SHE", "SHE", 2, 5),
        ("SHE", "SHE", 5, 5),
        ("POS", "SHE", 2, 2),
        ("POS", "SHE", 2, 5),
        ("POS", "SHE", 5, 2),
        ("POS", "SHE", 5, 5),
    }

    assert set(cls.keys()) == expected_keys


def test_mixing_matrices(data_path):
    mms = el.photo.mixing_matrices(data_path / "mixing_matrices.fits")

    expected_keys = {
        ("POS", "POS", 2, 2),
        ("POS", "POS", 2, 5),
        ("POS", "POS", 5, 5),
        ("SHE", "SHE", 2, 2),
        ("SHE", "SHE", 2, 5),
        ("SHE", "SHE", 5, 5),
        ("POS", "SHE", 2, 2),
        ("POS", "SHE", 2, 5),
        ("POS", "SHE", 5, 2),
        ("POS", "SHE", 5, 5),
    }

    assert set(mms.keys()) == expected_keys



__all__ = [
    "angular_power_spectra",
    "correlation_functions",
    "bandpowers",
    "cosebis",
    "mixing_matrices",
    "redshift_distributions",
]

from ._le3_pk_wl import (
    angular_power_spectra,
    mixing_matrices,
)

from ._le3_2pcf_wl import (
    correlation_functions,
    bandpowers,
    cosebis,
)

from ._phz import (
    redshift_distributions,
)
